@@ -15,7 +15,7 @@
  *  along with this program; see the file COPYING.  If not, write to
  *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
  *
- *  You may contact the author by e-mail:  hlub@knoware.nl
+ *  You may contact the author by e-mail:  hanslub42@gmail.com
  */
 
 #include "rlwrap.h"
@@ -23,6 +23,7 @@
 /* global vars */
 
 
+
 /* variables set via command line options */
 int always_readline = FALSE;	             /* -a option: always be in readline mode             */
 char *password_prompt_search_string = NULL;  /* (part of) password prompt (argument of -a option) */
@@ -52,9 +53,6 @@ char *filter_command = NULL;                 /* -z option: pipe prompts, input,
 
 /* variables for global bookkeeping */
 int master_pty_fd;		     /* master pty (rlwrap uses this to communicate with client) */
-int slave_pty_sensing_fd;		     /* slave pty (client uses this to communicate with rlwrap,
-				      * we keep it open after forking in order to keep track of
-				      * client's terminal settings */
 FILE *debug_fp = NULL;  	     /* filehandle of debugging log */
 char *program_name, *command_name;   /* "rlwrap" (or whatever has been symlinked to rlwrap)  and (base-)name of command */
 char *rlwrap_command_line = "";      /* rlwrap command line (rlwrap -options command <command_args> */
@@ -160,22 +158,25 @@ main(int argc, char **argv)
   char *command_name;
   rlwrap_command_line = unsplit_with(argc, argv, " ");     
   init_completer();
+
+  /* Harvest options and leave optind pointing to first non-option argument: */
   command_name = read_options_and_command_name(argc, argv);
-  command_line = unsplit_with(argc - optind, argv, " ");
-  
-  /* by now, optind points to <command>, and &argv[optind] is <command>'s argv */
+
+  /* by now, optind points to slave <command>, and &argv[optind] is <command>'s argv. Remember slave command line: */
+  command_line = unsplit_with(argc - optind, argv + optind, " ");
+
+  /* if stdin is not a tty, just execute <command>: */ 
   if (!isatty(STDIN_FILENO) && execvp(argv[optind], &argv[optind]) < 0)
-    /* if stdin is not a tty, just execute <command> */ 
     myerror(FATAL|USE_ERRNO, "Cannot execute %s", argv[optind]);	
   init_rlwrap(rlwrap_command_line);
   install_signal_handlers();	
   block_all_signals();
-  fork_child(command_name, argv);
+  fork_child(command_name, argv); /* this will unblock most signals most of the time */
   if (filter_command)
     spawn_filter(filter_command);
   
   main_loop();
-  return 42;			/* not reached, but some compilers are unhappy without this ... */
+  return 42;			/* The Answer, but, sadly, we'll never get there.... */
 }
 
 
@@ -185,7 +186,7 @@ main(int argc, char **argv)
  * read_options_and_command_name() has harvested rlwrap's own options
  */  
 static void
-fork_child(char *command_name, char **argv)
+fork_child(char *UNUSED(command_name), char **argv)
 {
   char *arg = argv[optind], *p;
   int pid;
@@ -251,7 +252,7 @@ main_loop()
   last_minute_checks();
   pass_through_filter(TAG_OUTPUT,""); /* If something is wrong with filter, get the error NOW */
   set_echo(FALSE);		/* This will also put the terminal in CBREAK mode */
-	test_main(); 
+  test_main(); 
   
   /* ------------------------------  main loop  -------------------------------*/
   while (TRUE) {
@@ -332,18 +333,12 @@ main_loop()
 	}	
 	if (!skip_rlwrap()) {                        /* ... or else, it is time to cook the prompt */
 	  if (pre_given && accepted_lines == 0) {
-	    saved_rl_state.input_buffer = mysavestring(pre_given); /* stuff pre-given text into edit buffer */
-	    saved_rl_state.point =  strlen(pre_given);
+            /* input_buffer and point have already been set in init_readline() */
 	    DPRINTF0(DEBUG_READLINE, "Starting line edit (because of -P option)");
 	    within_line_edit = TRUE;
 	    restore_rl_state();
-
 	    continue;
 	  }
-	  
-	  if (accepted_lines == 1 && one_shot_rlwrap) 
-	    cleanup_rlwrap_and_exit(EXIT_SUCCESS);
-
 			  
 	  move_cursor_to_start_of_prompt(ERASE); /* cooked prompt may be shorter than raw prompt, hence the ERASE */
 	  /* move and erase before cooking, as we need to move/erase according
@@ -503,7 +498,7 @@ main_loop()
 						   will make us read 2 bytes 0x90 and then 0xd7, (or maybe the other way round depending on endianness??)
 						   The readline library hides all this complexity and allows one to just "pass the bytes around" */
 	nread = read(STDIN_FILENO, &byte_read, 1);  /* read next byte of input   */
-	assert(sizeof(unsigned char) == 1);      /* gets optimised away */
+	assert(sizeof(unsigned char) == 1);         /* gets optimised away       */
 
 	if (nread <= 0) 
 	  DPRINTF1(DEBUG_TERMIO, "read from stdin returned %d", nread); 
@@ -533,9 +528,6 @@ main_loop()
 	    restore_rl_state();
 	  } 
 	                                        
-	  
-	  
-
 	  if (term_eof && byte_read == term_eof && strlen(rl_line_buffer) == 0) {	/* hand a term_eof (usually CTRL-D) directly to command */ 
 	    char *sent_EOF = mysavestring("?");
 	    *sent_EOF = term_eof;
@@ -547,7 +539,8 @@ main_loop()
 	    rl_stuff_char(byte_read);  /* stuff it back in readline's input queue */
 	    DPRINTF0(DEBUG_TERMIO, "passing it to readline");	
 	    DPRINTF2(DEBUG_READLINE, "rl_callback_read_char() (_rl_eof_char=%d, term_eof=%d)", _rl_eof_char, term_eof);
-	    rl_callback_read_char();
+            message_in_echo_area(NULL);     
+            rl_callback_read_char();
 	  }
 	}
       }
@@ -570,16 +563,10 @@ init_rlwrap(char *command_line)
 {
 
   char *homedir, *histdir, *homedir_prefix, *hostname;
+  struct stat statbuf;
   time_t now;
   
-  /* open debug file if necessary */
 
-  if (debug) {    
-    debug_fp = fopen(DEBUG_FILENAME, "w");
-    if (!debug_fp)
-      myerror(FATAL|USE_ERRNO, "Couldn't open debug file %s", DEBUG_FILENAME);
-    setbuf(debug_fp, NULL); /* always write debug messages to disk at once */
-  }
   hostname = getenv("HOSTNAME") ? getenv("HOSTNAME") : "?";
   now = time(NULL);
   DPRINTF0(DEBUG_ALL, "-*- mode: grep -*-");
@@ -604,12 +591,21 @@ init_rlwrap(char *command_line)
     histdir = homedir;
     history_filename = add3strings(homedir_prefix, command_name, "_history");
   }
+
+
   
   if (write_histfile) {
     if (access(history_filename, F_OK) == 0) {	/* already exists, can we read/write it? */
       if (access(history_filename, R_OK | W_OK) != 0) {
 	myerror(FATAL|USE_ERRNO, "cannot read and write %s", history_filename);
       }
+      /* OK, we can read and write it, but do we want to? Not if our effective UID
+        doesn't match the owner of the history file (like after sudo rlwrap on Ubuntu) */
+      assert(!stat(history_filename, &statbuf));
+      if(statbuf.st_uid != geteuid()) {
+        myerror(WARNING | NOERRNO, "Owner of %s and your effective UID don't match. History will be read-only", history_filename);
+        write_histfile = FALSE;
+      } 
     } else {			        /* doesn't exist, can we create it? */
       if(access(histdir, W_OK) != 0) {
         if (errno == ENOENT) {
@@ -622,11 +618,7 @@ init_rlwrap(char *command_line)
         }
       }
     }
-  } else {			/* ! write_histfile */
-    if (access(history_filename, R_OK) != 0) {
-      myerror(FATAL|USE_ERRNO, "cannot read %s", history_filename);
-    }
-  }
+  } 
 
   /* Initialize history */
   using_history();
@@ -664,6 +656,8 @@ init_rlwrap(char *command_line)
 static char *
 check_optarg(char opt, int remaining)
 {
+  MAYBE_UNUSED2(opt, remaining);
+  
   if (!optarg)
     last_option_didnt_have_optional_argument = TRUE; /* if this variable is set, and  if command is not found,
 							suggest that it may have been meant
@@ -743,24 +737,30 @@ read_options_and_command_name(int argc, char **argv)
       rl_basic_word_break_characters = add3strings("\r\n \t", optarg, "");
       opt_b = TRUE;
       break;
-    case 'c':	complete_filenames = TRUE; break;
+    case 'c':   complete_filenames = TRUE;
+#ifndef CAN_FOLLOW_COMMANDS_CWD
+                myerror(WARNING|NOERRNO, "On this system rlwrap cannot follow the rlwrapped command's working directory:\n"
+                                         "filename completion will only be relative to rlwrap's own working directory ");
+#endif
+                break;
     case 'C':	opt_C = mysavestring(optarg); break;
     case 'd':
 #ifdef DEBUG
       if (option_count > 1)
-        myerror(FATAL|NOERRNO, "-d or --debug option has to be the *first* rlwrap option");
-      if (check_optarg('d', remaining))
-        debug = atoi(optarg);
-      else
-        debug = DEBUG_DEFAULT;
+        myerror(FATAL|NOERRNO, "-d or --debug option has to be the *first* rlwrap option\n"
+                               "in order to be able to follow the processing  of all subsequent options");
+      debug = check_optarg('d', remaining) ? my_atoi(optarg) : DEBUG_DEFAULT;
+      debug_fp = fopen(DEBUG_FILENAME, "w");
+      if (!debug_fp)
+         myerror(FATAL|USE_ERRNO, "Couldn't open debug file %s", DEBUG_FILENAME);
+      setbuf(debug_fp, NULL); /* always write debug messages to disk at once */
 #else
-     myerror(FATAL|NOERRNO, "To use -d( for debugging), configure %s with --enable-debug and rebuild",
-         program_name);
+      myerror(FATAL|NOERRNO, "To use -d( for debugging), configure %s with --enable-debug and rebuild",program_name);
 #endif
       break;
 
     case 'D': 
-      history_duplicate_avoidance_policy=atoi(optarg);
+      history_duplicate_avoidance_policy=my_atoi(optarg);
       if (history_duplicate_avoidance_policy < 0 || history_duplicate_avoidance_policy > 2)
         myerror(FATAL|NOERRNO, "%s option with illegal value %d, should be 0, 1 or 2",
                 current_option('D', longindex), history_duplicate_avoidance_policy);
@@ -771,7 +771,7 @@ read_options_and_command_name(int argc, char **argv)
         myerror(FATAL|NOERRNO, "-e (--extra-char-after-completion) argument should be at most one character");
       break;
     case 'f':
-      if (strncmp(optarg, ".", 10) == 0)
+      if (strcmp(optarg, ".") == 0)
         feed_history_into_completion_list =  TRUE;
       else
         feed_file_into_completion_list(optarg);
@@ -799,10 +799,7 @@ read_options_and_command_name(int argc, char **argv)
     case 'M': multi_line_tmpfile_ext = mysavestring(optarg); break;
     case 'N': commands_children_not_wrapped = TRUE; break;
     case 'o': 
-      one_shot_rlwrap = TRUE;
-      impatient_prompt = FALSE;
-      wait_before_prompt = 200;
-      break;
+      one_shot_rlwrap = TRUE; break;
     case 'O': prompt_regexp = mysavestring(optarg); match_regexp("just testing", prompt_regexp, 1); break;
     case 'p':
       colour_the_prompt = TRUE;
@@ -819,8 +816,8 @@ read_options_and_command_name(int argc, char **argv)
     case 'r': remember_for_completion = TRUE;	break;
     case 'R': renice = TRUE;	break;
     case 's':
-      histsize = atoi(optarg);
-      if (histsize < 0) {
+      histsize = my_atoi(optarg);
+      if (histsize < 0 || *optarg == '-') {
         write_histfile = 0;
         histsize = -histsize;
       }
@@ -833,7 +830,7 @@ read_options_and_command_name(int argc, char **argv)
     case 'U': mirror_arguments = TRUE; break;
     case 'v': printf("rlwrap %s\n",  VERSION); exit(EXIT_SUCCESS);
     case 'w':
-      wait_before_prompt = atoi(optarg);
+      wait_before_prompt = my_atoi(optarg);
       if (wait_before_prompt < 0) {
         wait_before_prompt *= -1;
         impatient_prompt =  FALSE;
@@ -854,7 +851,7 @@ read_options_and_command_name(int argc, char **argv)
       usage(EXIT_FAILURE);
     }
   }
-
+  
   if (!complete_filenames && !opt_b) {	/* use / and . as default breaking characters whenever we don't complete filenames */
     rl_basic_word_break_characters =
       add2strings(rl_basic_word_break_characters, "/.");
@@ -872,7 +869,8 @@ read_options_and_command_name(int argc, char **argv)
     }
   }
   if (opt_C) {
-    int countback = atoi(opt_C);	/* investigate whether -C option argument is numeric */
+    int countback = 0;
+    if (isnumeric(opt_C)) countback = my_atoi(opt_C);	/* investigate whether -C option argument is numeric */
 
     if (countback > 0) {	/* e.g -C 1 or -C 12 */
       if (argc - countback < optind)	/* -C 666 */
@@ -952,7 +950,7 @@ flush_output_queue()
   how_much = min(BUFFSIZE, nl ? 1+ nl - output_queue : queuelen); /* never write more than one line, and never more than BUFFSIZE in one go */
   nwritten = write(master_pty_fd, output_queue, how_much);
 
-  assert(nwritten <= strlen(output_queue));
+  assert(nwritten <= (int) strlen(output_queue));
   if (debug) {
     char scratch = output_queue[nwritten];
     output_queue[nwritten] = '\0'; /* temporarily replace the last written byte + 1 by a '\0' */
@@ -1008,7 +1006,7 @@ cleanup_rlwrap_and_exit(int status)
   
   if (terminal_settings_saved)
     if (tcsetattr(STDIN_FILENO, TCSAFLUSH, &saved_terminal_settings) < 0)  /* ignore errors (almost dead anyway) */ 
-      ; /* fprintf(stderr, "Arggh\n"); don't use myerror!!*/
+      { /* nothing ... */ } /* fprintf(stderr, "Arggh\n"); don't use myerror!!*/
 
   if (!newline_came_last) /* print concluding newline, if necessary */
     my_putstr("\n");
@@ -1028,13 +1026,17 @@ cleanup_rlwrap_and_exit(int status)
 
 
 
+extern int test_haystack(const char *haystack, const char *needle);
 static void test_main() {
 #ifdef DEBUG
   if(debug & DEBUG_TEST_MAIN) {
     /* test, test */
-    test_unbackspace("abc","abc");
-    test_unbackspace("abc\bd","abd");
-    test_unbackspace("abc\r","abc");
+    test_haystack("rabarber","r");
+    test_haystack("","r");
+    test_haystack("rr rr rr rrr","rr");
+    /* test_haystack("bobobo",""); */
+    test_haystack("lalalalalal","la");
+
     test_unbackspace("abc\rd","dbc");
     test_unbackspace("abc\bd\re","ebd");
     test_unbackspace("abc\rde\rf","fec");
