@@ -16,7 +16,7 @@
     the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
     You may contact the author by:
-    e-mail:  hlub@knoware.nl
+    e-mail:  hanslub42@gmail.com
 */
 
 
@@ -26,7 +26,7 @@
 int remember_for_completion = FALSE;    /* whether we should put al words from in/output on the list */
 char *multiline_separator = NULL;       /* character sequence to use in lieu of newline when storing multi-line input in a single history line */
 char *pre_given = NULL;                 /* pre-given user input when rlwrap starts up */
-struct rl_state saved_rl_state = { "", "", 0, 0 };      /* saved state of readline */
+struct rl_state saved_rl_state = { "", "", 0, 0, 0 };      /* saved state of readline */
 static char return_key;                 /* Key pressed to enter line */
 static int forget_line;
 static char *colour_start, *colour_end;        /* colour codes */
@@ -37,27 +37,43 @@ static void line_handler(char *);
 static void my_add_history(char *);
 static int my_accept_line(int, int);
 static int my_accept_line_and_forget(int, int);
-static int dump_all_keybindings(int,int);
-static int munge_line_in_editor(int, int);
-static int please_update_alaf(int,int);
-static int please_update_ce(int,int);
-
+static void munge_file_in_editor(const char *filename, int lineno, int colno);
 static Keymap getmap(const char *name);
 static void bindkey(int key, rl_command_func_t *function, const char *maplist);
 
+
+/* readline bindable functions: */
+static int munge_line_in_editor(int, int);
+static int direct_keypress(int, int);
+static int handle_hotkey(int, int);
+static int handle_hotkey_without_history(int, int);
+static int please_update_alaf(int,int);
+static int please_update_ce(int,int);
+
+/* only useful while debugging: */
+static int debug_ad_hoc(int,int);
+static int dump_all_keybindings(int,int);
+
 void
-init_readline(char *prompt)
+init_readline(char *UNUSED(prompt))
 {
   DPRINTF1(DEBUG_READLINE, "Initialising readline version %x", rl_readline_version);   
   rl_add_defun("rlwrap-accept-line", my_accept_line,-1);
   rl_add_defun("rlwrap-accept-line-and-forget", my_accept_line_and_forget,-1);
-  rl_add_defun("rlwrap-dump-all-keybindings", dump_all_keybindings,-1);
   rl_add_defun("rlwrap-call-editor", munge_line_in_editor, -1);
+  rl_add_defun("rlwrap-direct-keypress", direct_keypress, -1);  
+  rl_add_defun("rlwrap-hotkey", handle_hotkey, -1);
+  rl_add_defun("rlwrap-hotkey-without-history", handle_hotkey_without_history, -1);
 
-  /* rlwrap bindable function names with underscores are deprecated: */
+  /* only useful while debugging */
+  rl_add_defun("rlwrap-dump-all-keybindings", dump_all_keybindings,-1);
+  rl_add_defun("rlwrap-debug-ad-hoc", debug_ad_hoc, -1);
+  
+  /* the old rlwrap bindable function names with underscores are deprecated: */
   rl_add_defun("rlwrap_accept_line_and_forget", please_update_alaf,-1);
   rl_add_defun("rlwrap_call_editor", please_update_ce,-1);
   
+  
   rl_variable_bind("blink-matching-paren","on"); /* Shouldn't this be on by default? */
 
   bindkey('\n', my_accept_line, "emacs-standard; vi-insert; vi-command"); 
@@ -65,10 +81,9 @@ init_readline(char *prompt)
   bindkey(15, my_accept_line_and_forget, "emacs-standard; vi-insert; vi-command");	/* ascii #15 (Control-O) is unused in readline's emacs and vi keymaps */
   if (multiline_separator) 
     bindkey(30, munge_line_in_editor, "emacs-standard;vi-insert;vi-command");            /* CTRL-^: unused in vi-insert-mode, hardly used in emacs  (doubles arrow-up) */
-  if (debug)
-    bindkey(7, dump_all_keybindings,"emacs-standard; vi-insert; vi-move; vi-command" /* "emacs-ctlx; emacs-meta" */); /* CTRL-G */
-  
 
+  
+  
   /* rl_variable_bind("gnah","gnerp"); It is not possible to create new readline variables (only functions) */
   rl_catch_signals = 0;
   rl_initialize();		/* This has to happen *before* we set rl_redisplay_function, otherwise
@@ -83,18 +98,17 @@ init_readline(char *prompt)
   
   rl_catch_signals = FALSE;
   rl_catch_sigwinch = FALSE;
-  saved_rl_state.input_buffer = mysavestring("");
+  saved_rl_state.input_buffer = mysavestring(pre_given ? pre_given : ""); /* Even though the pre-given input won't be displayed before the first 
+                                                                             cooking takes place, we still want it to be accepted when the user 
+                                                                             presses ENTER before that (e.g. because she already knows the 
+                                                                             pre-given input and wants to accept that) */
+  saved_rl_state.point = strlen(saved_rl_state.input_buffer);
   saved_rl_state.raw_prompt = mysavestring("");
   saved_rl_state.cooked_prompt = NULL;
   
 }
 
 
-
-
-
-
-
 /* save readline internal state in rl_state, redisplay the prompt
    (so that client output gets printed at the right place) */
 void
@@ -148,6 +162,24 @@ restore_rl_state()
   prompt_is_still_uncooked =  FALSE; /* has been done right now */
 }
 
+/* display (or remove, if message == NULL) message in echo area, with the appropriate bookkeeping */
+void
+message_in_echo_area(char *message)
+{
+  static int message_in_echo_area = FALSE;
+  DPRINTF1(DEBUG_READLINE, "message: %s", mangle_string_for_debug_log(message, MANGLE_LENGTH));
+  if (message) {
+    rl_save_prompt();
+    message_in_echo_area = TRUE;  
+    rl_message(message);
+  }  else {
+    if (message_in_echo_area)
+      rl_restore_prompt();
+    rl_clear_message();
+    message_in_echo_area = FALSE;
+  }     
+} 
+
 static void
 line_handler(char *line)
 {
@@ -222,8 +254,10 @@ line_handler(char *line)
         
     return_key = 0;
     within_line_edit = FALSE;
-
-    rl_callback_handler_remove();
+    if(!RL_ISSTATE(RL_STATE_MACROINPUT)) /* when called during playback of a multi-line macro, line_handler() will be called more 
+                                            than once whithout re-entering main_loop(). If we'd remove it here, the second call
+                                            would crash  */ 
+       rl_callback_handler_remove();
     set_echo(FALSE);
     free(saved_rl_state.input_buffer);
     free(saved_rl_state.raw_prompt);
@@ -235,6 +269,9 @@ line_handler(char *line)
     saved_rl_state.point = 0;
     saved_rl_state.already_saved = TRUE;
     redisplay  = TRUE;
+
+    if (one_shot_rlwrap)
+      write_EOF_to_master_pty();
   }
 }
 
@@ -242,7 +279,7 @@ line_handler(char *line)
 /* this function (drop-in replacement for readline's own accept-line())
    will be bound to RETURN key: */
 static int
-my_accept_line(int count, int key)
+my_accept_line(int UNUSED(count), int key)
 {
   rl_point = 0;			/* leave cursor on predictable place */
   my_redisplay();
@@ -253,7 +290,7 @@ my_accept_line(int count, int key)
 
 /* this function will be bound to rl_accept_key_and_forget key (normally CTRL-O) */
 static int
-my_accept_line_and_forget(int count, int key)
+my_accept_line_and_forget(int count, int UNUSED(key))
 {
   forget_line = 1;
   return my_accept_line(count, '\n');
@@ -268,6 +305,7 @@ dump_all_keybindings(int count, int key)
   return 0;
 }       
 
+
 /* format line and add it to history list, avoiding duplicates if necessary */
 static void
 my_add_history(char *line)
@@ -291,13 +329,12 @@ my_add_history(char *line)
   new_entry = filtered_line;    
   
   lookback = min(history_length, lookback);      
-  for (count = 0, here = history_length + history_base - 1;
+  for (count = 0, here = history_length - 1;
        count < lookback ;
        count++, here--) {
-    /* DPRINTF3(DEBUG_READLINE, "strcmping <%s> and <%s> (count = %d)", line, history_get(here)->line ,count); */
-    if (strncmp(new_entry, history_get(here) -> line, 10000) == 0) {
-      HIST_ENTRY *entry = remove_history (here - history_base); /* according to the history library doc this should be
-                                                                   remove_history(here) @@@?*/
+    DPRINTF4(DEBUG_READLINE, "strcmping <%s> and <%s> (count = %d, here = %d)", line, history_get(history_base + here)->line ,count, here);
+    if (strncmp(new_entry, history_get(history_base + here) -> line, 10000) == 0) { /* history_get uses the logical offset history_base .. */
+       HIST_ENTRY *entry = remove_history (here);                                   /* .. but remove_history doesn't!                      */
       DPRINTF2(DEBUG_READLINE, "removing duplicate entry #%d (%s)", here, entry->line);
       free_foreign(entry->line);
       free_foreign(entry);
@@ -468,44 +505,24 @@ my_redisplay()
 
 
 
-
-
-static int
-munge_line_in_editor(int count, int key)
+static void
+munge_file_in_editor(const char *filename, int lineno, int colno)
 {
-  int line_number = 0, column_number = 0,  ret, tmpfile_fd, bytes_read;
-  size_t tmpfilesize;
-  char *p, *tmpfilename, *text_to_edit;
-  char *editor_command1, *editor_command2, *editor_command3, *editor_command4,
-    *line_number_as_string, *column_number_as_string;
-  char *input, *rewritten_input, *rewritten_input2,  **possible_editor_commands;
 
-
-  if (!multiline_separator)
-    return 0;
-
-  tmpfile_fd = open_unique_tempfile(multi_line_tmpfile_ext, &tmpfilename);
-
-  text_to_edit =
-    search_and_replace(multiline_separator, "\n", rl_line_buffer, rl_point,
-                       &line_number, &column_number);
-  write_patiently(tmpfile_fd, text_to_edit, strlen(text_to_edit), "to temporary file");
-
-  if (close(tmpfile_fd) != 0) /* improbable */
-    myerror(FATAL|USE_ERRNO, "couldn't close temporary file %s", tmpfilename); 
+  int ret;
+  char *editor_command1, *editor_command2, *editor_command3,
+       *editor_command4, *line_number_as_string, *column_number_as_string,  **possible_editor_commands;
 
   /* find out which editor command we have to use */
   possible_editor_commands = list4(getenv("RLWRAP_EDITOR"), getenv("EDITOR"), getenv("VISUAL"), "vi +%L");
   editor_command1 = first_of(possible_editor_commands);
-  line_number_as_string = as_string(line_number);
-  column_number_as_string = as_string(column_number);
-  editor_command2 =
-    search_and_replace("%L", line_number_as_string, editor_command1, 0, NULL,
-                       NULL);
-  editor_command3 =
-    search_and_replace("%C", column_number_as_string, editor_command2, 0,
-                       NULL, NULL);
-  editor_command4 = add3strings(editor_command3, " ", tmpfilename);
+  line_number_as_string = as_string(lineno);
+  column_number_as_string = as_string(colno);
+  editor_command2 = search_and_replace("%L", line_number_as_string, editor_command1, 0, NULL, NULL);
+  editor_command3 = search_and_replace("%C", column_number_as_string, editor_command2, 0, NULL, NULL);
+  editor_command4 = strstr(editor_command3, "%F") ?
+    search_and_replace("%F", filename, editor_command3, 0, NULL, NULL) :
+    add3strings(editor_command3, " ", filename);
   
 
   /* call editor, temporarily restoring terminal settings */    
@@ -523,6 +540,35 @@ munge_line_in_editor(int count, int key)
   completely_mirror_slaves_terminal_settings();
   ignore_queued_input = TRUE;  
 
+  free_multiple(possible_editor_commands, editor_command2, editor_command3,
+                editor_command4, line_number_as_string, column_number_as_string, FMEND);
+}
+
+
+static int
+munge_line_in_editor(int UNUSED(count), int UNUSED(key))
+{
+  int line_number = 0, column_number = 0, tmpfile_fd, bytes_read;
+  size_t tmpfilesize;
+  char *p, *tmpfilename, *text_to_edit;
+  char *input, *rewritten_input, *rewritten_input2;
+
+
+  if (!multiline_separator)
+    return 0;
+
+  tmpfile_fd = open_unique_tempfile(multi_line_tmpfile_ext, &tmpfilename);
+
+  text_to_edit =
+    search_and_replace(multiline_separator, "\n", rl_line_buffer, rl_point,
+                       &line_number, &column_number);
+  write_patiently(tmpfile_fd, text_to_edit, strlen(text_to_edit), "to temporary file");
+
+  if (close(tmpfile_fd) != 0) /* improbable */
+    myerror(FATAL|USE_ERRNO, "couldn't close temporary file %s", tmpfilename); 
+
+  munge_file_in_editor(tmpfilename, line_number, column_number);
+  
   /* read back edited input, replacing real newline with substitute */
   tmpfile_fd = open(tmpfilename, O_RDONLY);
   if (tmpfile_fd < 0)
@@ -556,11 +602,7 @@ munge_line_in_editor(int count, int key)
   /* wash those dishes */
   if (unlink(tmpfilename))
     myerror(FATAL|USE_ERRNO, "could not delete temporary file %s", tmpfilename);
-  free(editor_command2);
-  free(editor_command3);
-  free(editor_command4);
-  free(line_number_as_string);
-  free(column_number_as_string);
+
   free(tmpfilename);
   free(text_to_edit);
   free(input);
@@ -571,6 +613,152 @@ munge_line_in_editor(int count, int key)
   return 0;
 }
 
+static int
+direct_keypress(int UNUSED(count), int key)
+{
+  char *key_as_str = mysavestring("?");
+  /* put the key in the output queue    */
+  *key_as_str = key;
+  DPRINTF1(DEBUG_READLINE,"direct keypress: %s", mangle_char_for_debug_log(key, TRUE));
+  put_in_output_queue(key_as_str);
+  free(key_as_str);
+  return 0;
+}
+
+static char* entire_history_as_one_string(void) {
+  HIST_ENTRY **the_list = history_list(), **entryp;
+  char *big_string = mymalloc(history_total_bytes() + history_length + 1);
+  char *stringp =  big_string;
+
+  if (!the_list) /* i.e. if there is no history */
+    return mysavestring("");
+
+  for (entryp = the_list; *entryp; entryp++) {
+    int length = strlen((*entryp)->line);
+    strncpy(stringp, (*entryp)->line, length); /* copy line, without closing NULL byte; */
+    stringp +=length;
+    *stringp++ = '\n';
+  }
+  *stringp = '\0';
+  DPRINTF1(DEBUG_READLINE, "stringified %d bytes of history", (int) strlen(big_string));
+  return big_string;
+}       
+
+static int
+debug_ad_hoc(int UNUSED(count), int UNUSED(hotkey))
+{
+  printf("\n%s", entire_history_as_one_string());
+  cleanup_rlwrap_and_exit(EXIT_SUCCESS);
+  return 42;
+}       
+  
+
+
+static int
+handle_hotkey2(int UNUSED(count), int hotkey, int without_history)
+{
+  char *prefix, *postfix, *history,  *histpos_as_string, *executing_keyseq;
+  char *new_prefix, *new_postfix, *new_history, *new_histpos_as_string, *message; 
+  char *filter_food, *filtered, **fragments,  *new_rl_line_buffer;
+  int nfilter_fields, length, new_histpos;
+  unsigned long int hash;
+
+  static const unsigned int MAX_HISTPOS_DIGITS = 6; /* one million history items should suffice */
+
+  
+#ifdef HAVE_RL_EXECUTING_KEYSEQ /* i.e. if readline version is >= 6.3 */
+  executing_keyseq = mysavestring(rl_executing_keyseq);
+#else
+  executing_keyseq = mysavestring("?");
+  *executing_keyseq = hotkey; /* The filter will only get the *last* byte of the key sequence that triggered rl_handle_hotkey */ 
+#endif
+
+    
+  DPRINTF3(DEBUG_READLINE, "hotkey press (without_history == %d): %x (%s)", without_history, hotkey, mangle_string_for_debug_log(executing_keyseq, MANGLE_LENGTH));
+
+  prefix = mysavestring(rl_line_buffer);
+  prefix[rl_point] = '\0';                                     /* chop off just before cursor */
+  postfix = mysavestring(rl_line_buffer + rl_point);
+
+  if (without_history) {
+    histpos_as_string = mysavestring("0");
+    history = mysavestring("");
+  } else {
+    histpos_as_string = as_string(where_history());
+    assert(strlen(histpos_as_string) <= MAX_HISTPOS_DIGITS);
+    history = entire_history_as_one_string();
+    hash = hash_multiple(2, history, histpos_as_string);
+  }     
+
+  /* filter_food = <keyseq_length><keyseq><prefix_length><prefix><postfix_length><postfix><history_length><history><histpos_length><histpos> + '\0' */
+  filter_food = merge_fields(executing_keyseq, prefix, postfix, history, histpos_as_string, END_FIELD); /* this is the format that the filter expects */
+
+  /* let the filter filter ...! */
+  filtered= pass_through_filter(TAG_HOTKEY, filter_food);
+  
+  /* OK, we now have to read back everything. After splitting the message, here should be exactly 5 components*/
+  fragments = split_filter_message(filtered, &nfilter_fields);
+  assert(nfilter_fields == 5); 
+  message               = fragments[0]; 
+  new_prefix            = fragments[1];
+  new_postfix           = fragments[2];
+  new_history           = fragments[3];
+  new_histpos_as_string = fragments[4];
+
+  if (!without_history && hash_multiple(2, new_history, new_histpos_as_string) != hash) { /* history has been rewritten */
+    char **linep, **history_lines = split_on_single_char(new_history, '\n', 0);
+    DPRINTF3(DEBUG_READLINE, "hash=%lx, new_history is %d bytes long, histpos <%s>", hash, (int) strlen(new_history), new_histpos_as_string);
+    clear_history();
+    for (linep = history_lines; *linep; linep++) 
+      add_history(*linep);
+    new_histpos = my_atoi(new_histpos_as_string);
+    history_set_pos(new_histpos);
+    free_splitlist(history_lines);
+  }
+  new_rl_line_buffer = add2strings(new_prefix, new_postfix);
+
+  if ( (length = strlen(new_rl_line_buffer)) > 0  &&  new_rl_line_buffer[length - 1] == '\n') {
+    new_rl_line_buffer[length - 1] = '\0';
+    rl_done = TRUE;
+    return_key = (char) '\n';
+    assert(strchr(new_rl_line_buffer, '\n') == NULL);
+  }
+
+  rl_delete_text(0, strlen(rl_line_buffer));
+  rl_point = 0;
+  rl_insert_text(new_rl_line_buffer);
+  rl_point = strlen(new_rl_line_buffer);
+
+  
+  
+  if (*message && *message != hotkey) {                          /* if message has been set (i.e. != hotkey) , and isn't empty: */ 
+    message = append_and_free_old(mysavestring(message), " ");   /* put space (for readability) between the message and the input line .. */
+    message_in_echo_area(message);                          /* .. then write it to echo area */
+  }     
+
+  clear_line();
+  rl_on_new_line();
+  rl_redisplay();
+ 
+
+  free_splitlist(fragments);                                   /* this will free all the fragments (and the list itself) in one go  */
+  free_multiple(prefix, postfix, filter_food, executing_keyseq, filtered, new_rl_line_buffer, history, histpos_as_string, FMEND);
+  return 0;
+}
+
+
+static int
+handle_hotkey(int count, int hotkey)
+{
+  return handle_hotkey2(count, hotkey, FALSE);
+}       
+
+
+static int
+handle_hotkey_without_history(int count, int hotkey)
+{
+  return handle_hotkey2(count, hotkey, TRUE);
+}       
 
 
 void
@@ -674,7 +862,7 @@ prompt_is_single_line(void)
 }
 
 
-char *process_new_output(const char* buffer, struct rl_state* state) {
+char *process_new_output(const char* buffer, struct rl_state* UNUSED(state)) {
   char*last_nl, *old_prompt_plus_new_output, *new_prompt, *result;
   
   old_prompt_plus_new_output = append_and_free_old(saved_rl_state.raw_prompt, buffer); 
@@ -785,11 +973,11 @@ static int please_update(const char *varname) {
   return 0;
 }       
 
-static int please_update_alaf(int count, int key) {
+static int please_update_alaf(int UNUSED(count), int UNUSED(key)) {
   return please_update("rlwrap_accept_line_and_forget");
 }
 
-static int please_update_ce(int count, int key) {
+static int please_update_ce(int UNUSED(count), int UNUSED(key)) {
   return please_update("rlwrap_call_editor");
 }
 
