@@ -1,3 +1,4 @@
+
 /*  rlwrap.h: includes, definitions, declarations */
 
 /*  This program is free software; you can redistribute it and/or modify
@@ -15,7 +16,7 @@
     the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
     You may contact the author by:
-    e-mail:  hlub@knoware.nl
+    e-mail:  hanslub42@gmail.com
 */
 
 #include "../config.h"
@@ -71,6 +72,9 @@
 #endif
 
 
+#ifdef HAVE_LIBGEN_H
+#  include <libgen.h>
+#endif
 
 #ifdef HAVE_CURSES_H
 #  include <curses.h>
@@ -117,7 +121,21 @@ char *strchr(), *strrchr();
 #endif
 
 
+#if HAVE_DECL_PROC_PIDVNODEPATHINFO
+#  include <libproc.h>
+#endif
 
+#if HAVE_FREEBSD_LIBPROCSTAT
+#  include <sys/param.h>
+#  include <sys/queue.h>
+#  include <sys/socket.h>
+#  include <libprocstat.h>
+#  include <sys/sysctl.h>
+#endif
+
+#if HAVE_PROC_PID_CWD || HAVE_DECL_PROC_PIDVNODEPATHINFO || HAVE_FREEBSD_LIBPROCSTAT
+#  define CAN_FOLLOW_COMMANDS_CWD 1
+#endif
 
 #define BUFFSIZE 512
 
@@ -246,6 +264,7 @@ extern struct rl_state
 
 void save_rl_state(void);
 void restore_rl_state(void);
+void message_in_echo_area(char *message);
 void init_readline(char *);
 void my_redisplay(void);
 void initialise_colour_codes(char *colour);
@@ -318,10 +337,12 @@ void  mirror_args(pid_t command_pid);
 #define USE_ERRNO 1
 #define NOERRNO   0
 
+/* constant to signify the end of a free_multiple() argument list (NULL would't work) */
+#define FMEND  ((void *) -1)
 
 void  myerror(int error_flags, const char *message, ...);
 void  *mymalloc(size_t size);
-void  myfree(void *ptr);
+void  free_multiple(void *ptr, ...);
 void  mysetsid(void);
 void  close_open_files_without_writing_buffers(void);
 size_t filesize(const char *filename);
@@ -329,6 +350,7 @@ void  open_logfile(const char *filename);
 void  write_logfile(const char *str);
 void  close_logfile(void);
 void  timestamp(char *buf, int size);
+unsigned long hash_multiple(int n, ...);
 int   killed_by(int status);
 void  change_working_directory(void);
 void  log_terminal_settings(struct termios *terminal_settings);
@@ -336,22 +358,22 @@ void  log_fd_info(int fd);
 void  last_minute_checks(void);
 void  mymicrosleep(int msec);
 void  do_nothing(int unused);
-extern char slaves_working_directory[];
 
 
 /* in string_utils.c */
-char *mybasename(char *filename);
-char *mydirname(char *filename);
+char *mybasename(const char *filename);
+char *mydirname(const char *filename);
 void  mystrlcpy(char *dst, const char *src, size_t size);
 void  mystrlcat(char *dst, const char *src, size_t size);
 char *mystrstr(const char *haystack, const char *needle);
 char *mysavestring(const char *string);
 char *add3strings(const char *str1, const char *str2, const char *str3);
 #define add2strings(a,b)  add3strings(a,b,"")
+int my_atoi(const char *nptr);
 char *mystrtok(const char *s, const char *delim);
 char **split_with(const char *string, const char *delim);
 char *unsplit_with(int n, char ** strings, const char *delim);
-char **split_on_single_char(const char *string, char c);
+char **split_on_single_char(const char *string, char c, int expected_count);
 int scan_metacharacters(const char* string, const char *metacharacters);
 char **list4 (char *el0, char *el1, char *el2, char *el3);
 void free_splitlist (char **list);
@@ -360,7 +382,7 @@ char *mangle_char_for_debug_log(char c, int quote_me);
 char *mangle_string_for_debug_log(const char *string, int maxlen);
 char *mangle_buffer_for_debug_log(const char *buffer, int length);
 char *mem2str(const char *mem, int size);
-char *search_and_replace(char *patt, char *repl, const char *string,
+char *search_and_replace(char *patt, const char *repl, const char *string,
                          int cursorpos, int *line, int *col);
 char *first_of(char **strings);
 char *as_string(int i);
@@ -376,7 +398,11 @@ char *get_last_screenline(char *long_line, int termwidth);
 char *lowercase(const char *str);
 char *colour_name_to_ansi_code(const char *colour_name);
 int match_regexp(const char *string, const char *regexp, int case_insensitive);
-
+int isnumeric(char *string);
+#define END_FIELD (char*)NULL /* marker object to terminate vargs */
+char *append_field_and_free_old(char *message, const char *field);
+char *merge_fields(char *field, ...);
+char **split_filter_message(char *message, int *count);
 
 /* in pty.c: */
 pid_t my_pty_fork(int *, const struct termios *, const struct winsize *);
@@ -434,11 +460,17 @@ extern int newline_came_last;
 
 /* in filter.c */
 
+#define MAX_TAG 255
 #define TAG_INPUT 0
 #define TAG_OUTPUT 1
 #define TAG_HISTORY 2
 #define TAG_COMPLETION 3
 #define TAG_PROMPT 4
+#define TAG_HOTKEY 5
+#define MAX_INTERESTING_TAG 5 /* max tag for which the filter can have a handler */
+
+#define TAG_WHAT_ARE_YOUR_INTERESTS 127
+
 
 #define TAG_IGNORE 251
 #define TAG_ADD_TO_COMPLETION_LIST 252
@@ -512,11 +544,17 @@ void filter_test(void);
 #  define DEBUG_DEFAULT                          (DEBUG_TERMIO | DEBUG_SIGNALS | DEBUG_READLINE)
 #  define DEBUG_ALL                              (2*DEBUG_MAX-1)
 
+# define MAYBE_UNUSED(x)                         (void) (x)
+# define MAYBE_UNUSED2(x,y)                      (void) (x); (void) (y)
 
-
-
-# ifndef __GNUC__
-# define __FUNCTION__ ""
+# ifdef __GNUC__
+#   define __MYFUNCTION__ __extension__ __FUNCTION__
+#   define UNUSED(x) UNUSED_ ## x __attribute__((__unused__))
+#   define UNUSED_FUNCTION(f) __attribute__((__unused__)) UNUSED_ ## f
+# else
+#   define __MYFUNCTION__ ""
+#   define UNUSED(x) UNUSED_ ## x
+#   define UNUSED_FUNCTION(f) UNUSED_ ## f  
 # endif
 
 # define WHERE_AND_WHEN \
@@ -524,7 +562,7 @@ void filter_test(void);
   if(debug & DEBUG_WITH_TIMESTAMPS) timestamp(when, sizeof(when)); else *when='\0'; \
   debug = 0; /* don't debug while evaluating the DPRINTF arguments */ \
   snprintf2(file_line, sizeof(file_line),"%.15s:%d:",__FILE__,__LINE__); \
-  fprintf(debug_fp, "%-20s %s %-25.25s ", file_line, when, __FUNCTION__);\
+  fprintf(debug_fp, "%-20s %s %-25.25s ", file_line, when, __MYFUNCTION__);\
 
 
 #  define NL_AND_FLUSH           fputc('\n', debug_fp) ; fflush(debug_fp); debug = debug_saved;
@@ -552,9 +590,13 @@ void filter_test(void);
 
 
 #else
-#  define HEAVEN_FORBID(exp)     (exp)
-#  define OH_PLEASE(exp)         (exp)
+#  define UNUSED(x) x
+#  define UNUSED_FUNCTION(f) f
+#  define MAYBE_UNUSED(x)
+#  define MAYBE_UNUSED2(x,y)
 #  define MANGLE_LENGTH          0
+#  define MAYBE_UNUSED(x)           
+#  define MAYBE_UNUSED2(x,y)                     
 #  define DPRINTF0(mask, format)
 #  define DPRINTF1(mask, format,arg)
 #  define DPRINTF2(mask, format,arg1, arg2)
@@ -562,8 +604,11 @@ void filter_test(void);
 #  define DPRINTF4(mask, format,arg1, arg2, arg3, arg4)
 #  define ERRMSG(b)
 #  define SHOWCURSOR
-#  define NDEBUG                /* disable assertions */
 #  define DEBUG_RANDOM_SLEEP  
+
+/* Don't #define NDEBUG!  There are assertions that cannot be skipped, as in assert(important_function_call()) */
+/* Todo (maybe) #define myassert(x) if(DEBUG){assert(x)} for the other (skippable) assertions                  */
+
 #endif
 
 #include <assert.h>
