@@ -133,10 +133,15 @@ append_and_free_old(char *str1, const char *str2)
 /* mybasename and mydirname: wrappers around basename and dirname, if
    available, otherwise emulations of them */ 
 
-char *
-mybasename(char *filename)
-{                               /* determine basename of "filename" */
 
+char *
+mybasename(const char *filename)
+{                               /* determine basename of "filename" */
+#if defined(HAVE_BASENAME) && defined(_GNU_SOURCE) /* we want only the GNU version */
+  char *filename_copy = mysavestring(filename);
+  char *result = basename(filename_copy);
+  return result;
+#else
   char *p;
 
   /* find last '/' in name (if any) */
@@ -144,11 +149,17 @@ mybasename(char *filename)
     if (*(p - 1) == '/')
       break;
   return p;
+#endif
 }
 
 char *
-mydirname(char *filename)
+mydirname(const char *filename)
 {                               /* determine directory component of "name" */
+#ifdef HAVE_DIRNAME
+  char *filename_copy = mysavestring(filename);
+  char *result = dirname(filename_copy);
+  return result;
+#else
   char *p;
 
   /* find last '/' in name (if any) */
@@ -156,30 +167,89 @@ mydirname(char *filename)
     if (*(p - 1) == '/')
       break;
   return (p == filename ? "." : mystrndup(filename, p - filename));
+#endif
 }
 
 
-/* mystrtok: saner version of strtok that doesn't overwrite its first argument */
+
+/* Better atoi() with error checking */
+int
+my_atoi(const char *nptr)
+{
+  int result;
+  char *endptr;
+  
+  errno = 0;
+  result = (int) strtol(nptr, &endptr, 10);
+  if (errno || endptr == nptr || *endptr)
+    myerror(FATAL|USE_ERRNO, "Could not make sense of <%s> as an integer", mangle_string_for_debug_log(nptr, 10));
+  return result;
+}       
+
+/* TODO: clean up the following mess. strtok() is cute, but madness. Write one function
+   char *tokenize(const char *string, const char *delimiters, bool allow_empty_strings), and make 
+   both split_with functions a special case of it. Drop mystrtok, count_str_occurrences and count_char_occurrences */ 
+
+
+  /* mystrtok: saner version of strtok that doesn't overwrite its first argument */
+  /* Scary strtok: "The  strtok()  function breaks a string into a sequence of zero or more nonempty tokens.  
+    On the first call to strtok(), the string to be parsed should be specified in str.  
+    In each subsequent call that should parse the same string, str must be NULL.
+  */
+
 
 char *
-mystrtok(const char *s, const char *delim) {
+mystrtok(const char *s, const char *delimiters) {
   static char *scratchpad = NULL;
-  if (s) {
-    if (scratchpad)
-      free(scratchpad);
+  if (s) { /* first call */
+    if (scratchpad) 
+      free(scratchpad); /* old news */
     scratchpad = mysavestring(s);
-  }    
-  return strtok(s ? scratchpad : NULL, delim);
+  }
+  return strtok(s ? scratchpad : NULL, delimiters);
 }       
 
+
+static int
+count_str_occurrences(const char *haystack, const char* needle)
+{
+  int count = 0, needle_length = strlen(needle);
+  const char *p = haystack;
+  
+  assert(needle_length > 0);
+  while ((p = strstr(p, needle))) {
+      count++;
+      p += needle_length;
+  }
+  return count;
+}
+
+static int
+count_char_occurrences(const char *haystack, char c)
+{
+  int count;
+  char *needle = mysavestring(" ");
+  *needle = c;
+  count = count_str_occurrences(haystack, needle);
+  free(needle);
+  return count;
+}
+
+
+void test_haystack(const char *haystack, const char* needle) {
+  printf("<%s> contains <%s> %d times\n", haystack, needle, count_str_occurrences(haystack, needle));
+}       
+
+
+
 /* split_with("a bee    cee"," ") returns a pointer to an array {"a", "bee",  "cee", NULL} on the heap */
 
 char **
-split_with(const char *string, const char *delim) {
+split_with(const char *string, const char *delimiters) {
   const char *s;
   char *token, **pword;
-  char **list = mymalloc(1 + strlen(string) * sizeof(char **)); /* list can never be longer than string + 1 */ 
-  for (s = string, pword = list; (token = mystrtok(s, delim)); s = NULL) 
+  char **list = mymalloc((1 + strlen(string)) * sizeof(char **)); /* worst case: only delimiters  */ 
+  for (s = string, pword = list; (token = mystrtok(s, delimiters)); s = NULL) 
     *pword++ = mysavestring(token);
   *pword = NULL;
   return list;
@@ -197,10 +267,11 @@ unsplit_with(int n, char **strings, const char *delim) {
   return result;
 }
 
-/* split_with("a\t\tbla","\t") returns {"a" "bla", NULL}, but we want {"a", "", "bla", NULL} for filter completion.
-   We write a special version (can be freed with free_splitlist) */
-char **split_on_single_char(const char *string, char c) {
-  char **list = mymalloc(1 + strlen(string) * sizeof(char **));
+/* split_with("a\t\tbla", '\t') returns {"a" "bla", NULL}, but we want {"a", "", "bla", NULL} for filter completion.
+   We write a special version (can be freed with free_splitlist), that optionally checkst the number of components (if expected_count > 0) */
+char **split_on_single_char(const char *string, char c, int expected_count) {
+  /* the 1st +1 for the last element ("bla"), the 2nd +1 for the marker element (NULL) */
+  char **list = mymalloc((count_char_occurrences(string,c) + 1 + 1) * sizeof(char **));
   char *stringcopy = mysavestring(string);
   char *p, **pword, *current_word;
   
@@ -213,6 +284,9 @@ char **split_on_single_char(const char *string, char c) {
     }
   }
   *pword++ = mysavestring(current_word);
+  if (expected_count  && pword-list != expected_count) 
+    myerror(FATAL|NOERRNO, "splitting <%s> on single %s yields %d components, expected %d",
+            mangle_string_for_debug_log(string, 30), mangle_char_for_debug_log(c, 1), pword -  list, expected_count); 
   *pword = NULL;
   free(stringcopy);
   return list;
@@ -246,7 +320,7 @@ void free_splitlist (char **list) {
    
 
 char *
-search_and_replace(char *patt, char *repl, const char *string, int cursorpos,
+search_and_replace(char *patt, const char *repl, const char *string, int cursorpos,
                    int *line, int *col)
 {
   int i, j, k;
@@ -429,6 +503,7 @@ list4 (char *el0, char *el1, char *el2, char *el3)
   list[1] = el1;
   list[2] = el2;
   list[3] = el3;
+  DPRINTF4(DEBUG_AD_HOC, "el0: <%s> el1: <%s> el2: <%s> el3: <%s>", el0, el1, el2, el3);
   return list;
 }
 
@@ -454,7 +529,7 @@ void remove_padding_and_terminate(char *buf, int length) {
       *copyptr++ = *readptr;
   }
   *copyptr = '\0';
-  if (debug && strlen(buf) != length)
+  if (debug && strlen(buf) != (unsigned int) length)
     DPRINTF2(DEBUG_TERMIO, "removed %d zero bytes (padding?) from %s", length - (int) strlen(buf), mangle_string_for_debug_log(buf, MANGLE_LENGTH));
 }       
         
@@ -822,3 +897,160 @@ int match_regexp (const char *string, const char *regexp, int case_insensitive)
   
   return result;        
 }
+
+
+/* returns TRUE if string is numeric (i.e. positive or negative integer), otherwise FALSE */
+int isnumeric(char *string){
+  char *pstr = string;
+  if (*pstr == '-')  /* allow negative numbers */
+    pstr++; 
+  while (*pstr != '\0')
+    if (!isdigit(*pstr++)) return FALSE;
+
+  return TRUE;
+}
+
+#define DIGITS_NUMBER 8  /* number of digits of length of field */
+#define MAX_FIELD_LENGTH (~((-1)<<(DIGITS_NUMBER * 4 -1)) -1)
+#define MY_HEX_FORMAT(n) "%0" MY_ITOA(n) "x"
+#define MY_ITOA(n) #n
+
+
+/* fussy strtol with error checking */
+long
+mystrtol(const char *nptr, int base)
+{
+  char *endptr;
+  long result;
+
+  errno = 0;
+  result = strtol(nptr, &endptr, base);
+  if (*endptr != '\0')
+    myerror(FATAL|NOERRNO, "invalid representation %s", nptr);
+  if (errno != 0)
+    myerror(FATAL|USE_ERRNO, "strtol error");
+
+  return result;
+}
+
+
+
+
+/* Encode a length as a string on the heap */
+static char *
+encode_field_length(int length)
+{
+  char *encoded_length = mymalloc(DIGITS_NUMBER+1);
+  sprintf(encoded_length, MY_HEX_FORMAT(DIGITS_NUMBER), length);
+  return encoded_length;
+}       
+
+
+/* decode first length field in a message of form "<length 1> <message 1> <length 2> ...." 
+   and advance pointer *ppmessage to the start of <message 1> */
+static int
+decode_field_length(char** ppmessage)
+{
+  char hex_string[DIGITS_NUMBER+1];
+  long length;
+  
+  mystrlcpy(hex_string, *ppmessage, DIGITS_NUMBER+1);
+  length = mystrtol(hex_string, 16);
+  *ppmessage += DIGITS_NUMBER;
+  return length;
+}       
+  
+
+/* Test an invariant: */
+void
+test_field_length_encoding()
+{
+  int testval = 1423722;
+  char *encoded = encode_field_length(testval);
+  assert( decode_field_length(&encoded) == testval &&  (*encoded == '\0'));
+}
+
+/* Append <field_length> <field> to message and return the result (after freeing the original message)
+<field_length> is a string representation of <field>s length
+message can be empty, or, equivalently, NULL.
+*/
+char *
+append_field_and_free_old(char *message, const char *field)
+{
+  int length = strlen(field);
+  char *encoded_length;
+  if (length > MAX_FIELD_LENGTH)
+    myerror(FATAL|NOERRNO, "message field\"%s...\" has length %d, it should be less than %d",
+            mangle_string_for_debug_log(field, 10), length,  MAX_FIELD_LENGTH);
+  encoded_length =  encode_field_length(length);
+  message = append_and_free_old(message, encoded_length);
+  message = append_and_free_old(message, field);
+  free(encoded_length);
+  return message;
+}
+
+
+
+char *
+merge_fields(char *field, ...)
+{  
+  char *varg = field;
+  char *message = NULL;
+  va_list vargs;
+  
+  va_start(vargs, field);
+  while (varg != END_FIELD) {
+    message = append_field_and_free_old(message, varg);
+    varg = va_arg(vargs, char*);
+  }
+  va_end(vargs);
+  return message;
+}
+
+
+
+/*
+split a message of a string:
+
+    <field_length1> <field1> <field_length 2> <field2>...
+
+into:
+
+    [<field1>, <field2>, ...]
+*/
+
+
+char **
+split_filter_message(char *message, int *counter)
+{
+  char *pmessage = message;
+  int message_length = strlen(message);
+  char **list, **plist;
+  int nfields = 0;
+
+  static int smallest_message_size = 0;
+  
+  if (smallest_message_size == 0)
+    smallest_message_size = strlen(append_field_and_free_old(NULL, "")); /* this assumes that the empty message is the smallest possible */
+
+  assert(smallest_message_size > 0);
+  plist = list = mymalloc(sizeof(char*) * (1 + strlen(message)/smallest_message_size )); /* worst case: "0000000000000000000000" */
+  nfields = 0;
+
+  while(!(*pmessage == '\0')) {
+    long length = decode_field_length(&pmessage);
+
+    /* cut out a field from the head of the message: */
+    char *field = mymalloc(sizeof(char) * (length+1));
+    mystrlcpy(field, pmessage, length+1);
+    *plist++ = field;
+    pmessage += length;
+    nfields++;
+    if (pmessage > message + message_length)
+      myerror(FATAL|NOERRNO, "malformed message; %s", mangle_string_for_debug_log(message, 256));
+  }
+  if (counter)
+    *counter =  nfields;
+  *plist = 0;
+  return list;
+}
