@@ -16,21 +16,29 @@
     the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
     You may contact the author by:
-    e-mail:  hlub@knoware.nl
+    e-mail:  hanslub42@gmail.com
 */
 
 
 #include "rlwrap.h"
 
+/* rlwrap needs a little bit of cursor acrobatics to cope with multi-line prompts, recovering from CTRL-Z and editors that may have used 
+the alternate screen. All such cursor movements are done via the functions in this file. In the past, rlwrap relied on only the termcap 
+library (or an emulation of it), and would fall back on reasonable defaults ("\r", "\b" etc) if this was not found. 
+
+In the future it will also use terminfo or even ncurses. But the fall-back to older libraries is still important, as rlwrap seems 
+to be especially needed on older machines                                                                                            */ 
+
 /*global vars */
 char term_eof;                  /* end_of_file char */
 char term_stop;                 /* stop (suspend) key */
 char *term_backspace;           /* backspace control seq  (or 0, if none defined in terminfo) */
-char *term_cursor_hpos;
+char *term_cursor_hpos;         /* control seq to position cursor st given position on current line */ 
 char *term_clear_screen;
 char *term_cursor_up;
 char *term_cursor_down;
 char *term_cursor_left;         /* only used for debugging (the SHOWCURSOR macro)                */
+char *term_cursor_right;        /* only used to emulate a missing term_cursor_hpos               */
 char *term_rmcup;               /* rmcup - char sequence to return from alternate screen         */
 char *term_rmkx;                /* rmkx - char sequence to return from keyboard application mode */
 /* int term_colors = -1;        number of colors ('colors' capability, only in terminfo)         */
@@ -46,24 +54,54 @@ static char *term_cr;           /* carriage return (or 0, if none defined in ter
 static char *term_clear_line;
 char *term_name;
 
-/* TODO: replace termcap functions by terminfo ones, keeping termcap as a fallback for ancient systems
-   we might have to give both terminfo-style and termcap names, like in my_tigetstr("cub1", "le") */
 
 static char *my_tgetstr (char *id) {
+#ifdef HAVE_TERMCAP_H
   char *term_string_buf = (char *)mymalloc(2048), *tb = term_string_buf;
   char *stringcap = tgetstr(id, &tb); /*  rl_get_termcap(id) only gets capabilities used by readline */
   char *retval = stringcap ? mysavestring(stringcap) : NULL; 
   DPRINTF2(DEBUG_TERMIO, "tgetstr(\"%s\") = %s", id, (stringcap ? mangle_string_for_debug_log(stringcap,20) : "NULL"));
   free(term_string_buf);
   return retval;
+#else
+  return NULL;
+#endif
 }
 
-#define FALLBACK_TERMINAL "vt100" /* terminal to use when we don't know which terminal we're on */
+static char *my_tigetstr (char *tid) {
+#ifdef HAVE_CURSES_H
+  static int term_has_been_setup = FALSE;
+  char *stringcap, *retval;
+  if (!term_has_been_setup) { 
+     setupterm(term_name, STDIN_FILENO, (int *)0); /* like tgetent() before tgetstr(), we have to call setupterm() befor tigetstr() */  
+     term_has_been_setup = TRUE;
+  }     
+  stringcap = tigetstr(tid); 
+  retval = stringcap ? mysavestring(stringcap) : NULL;
+  DPRINTF2(DEBUG_TERMIO, "tigetstr(\"%s\") = %s", tid, (stringcap ? mangle_string_for_debug_log(stringcap,20) : "NULL"));
+  return retval;
+#else
+  return NULL;
+#endif
+}
+
+/* Get (copies of) escape codes ("string capabilities") by name. First try terminfo name, then termcap name  */
+static char *
+tigetstr_or_else_tgetstr(char *capname, char *tcap_code)
+{
+  char *retval;
+  retval = my_tigetstr(capname);
+  if (!retval)
+    retval = my_tgetstr(tcap_code);
+  return retval;
+}
+  
+#define FALLBACK_TERMINAL "vt100" /* hard-coded terminal name to use when we don't know which terminal we're on */
 
 void
 init_terminal(void)
 {                               /* save term settings and determine term type */
-  char *term_buf;
+  char *term_buf = mymalloc(2048);
   int  we_have_stringcaps;
   int  we_are_on_hp_ux11;
 
@@ -83,61 +121,65 @@ init_terminal(void)
   /* init some variables: */
   term_name = getenv("TERM");
   DPRINTF1(DEBUG_TERMIO, "found TERM = %s", term_name);  
+  
+  
 
+
+
+
+  /* On weird and scary HP-UX 11 succesful tgetent() returns 0, so we need the followig to portably catch errors: */  
+  we_are_on_hp_ux11 = tgetent(term_buf, "vt100") == 0 && tgetent(term_buf,"QzBgt57gr6xwxw") == -1; /* assuming there is no terminal called Qz... */
+  #define T_OK(retval) ((retval) > 0 || ((retval)== 0 && we_are_on_hp_ux11))
+
+  /* if TERM is not set, use FALLBACK_TERMINAL */
   if (!term_name || strlen(term_name)==0) {
     myerror(WARNING|USE_ERRNO, "environment variable TERM not set, assuming %s", FALLBACK_TERMINAL); 
     term_name = FALLBACK_TERMINAL;
   }
-   
-  term_buf = (char *)mymalloc(4096);
-
-  term_backspace = NULL;
-  term_cr = NULL;
-  term_clear_line = NULL;
-  term_cursor_hpos = NULL;
-   
-  /* On weird and scary HP-UX 11 succesful tgetent() returns 0: */  
-  we_are_on_hp_ux11 = tgetent(term_buf, "vt100") == 0 && tgetent(term_buf,"QzBgt57gr6xwxw") == -1; /* assuming there is no terminal called Qz... */
-  #define T_OK(retval) ((retval) > 0 || ((retval)== 0 && we_are_on_hp_ux11))
-
-  /* Test whether we can find stringcaps, use FALLBACK_TERMINAL name if not  */
+  
+  /* If we cannot find stringcaps, use FALLBACK_TERMINAL  */
   if (! (we_have_stringcaps = T_OK(tgetent(term_buf, term_name))) && strcmp(term_name, FALLBACK_TERMINAL)) { 
     myerror(WARNING|NOERRNO, "your $TERM is '%s' but %s couldn't find it in the terminfo database. We'll use '%s'", 
                             term_name, program_name, FALLBACK_TERMINAL);
     term_name = FALLBACK_TERMINAL;
-  }       
-  if (! (we_have_stringcaps = T_OK(tgetent(term_buf, FALLBACK_TERMINAL)))) 
-    myerror(WARNING|NOERRNO, "Even %s is not found in the terminfo database. Expect some problems...", FALLBACK_TERMINAL);
-  
-  
+  }
+
+
+  /* If we cannot find stringcaps, even for FALLBACK_TERMINAL, complain */
+  if (!we_have_stringcaps && ! (we_have_stringcaps = T_OK(tgetent(term_buf, FALLBACK_TERMINAL)))) 
+    myerror(WARNING|NOERRNO, "Even %s is not found in the terminfo database. Expect some problems.", FALLBACK_TERMINAL);
+
   DPRINTF1(DEBUG_TERMIO, "using TERM = %s", term_name);  
   mysetenv("TERM", term_name);
 
+ 
+
+  
   if (we_have_stringcaps)  { 
-    term_backspace      = my_tgetstr("le");
-    term_cr             = my_tgetstr("cr");
-    term_clear_line     = my_tgetstr("ce");
-    term_clear_screen   = my_tgetstr("cl");
-    term_cursor_hpos    = my_tgetstr("ch");
-    term_cursor_left    = my_tgetstr("le");
-    term_rmcup          = my_tgetstr("te"); /* rlwrap still uses ye olde termcappe names */
-    term_rmkx           = my_tgetstr("ke");
-    if (term_cursor_hpos && !strstr(term_cursor_hpos, "%p")) {
-      /* Some RedHat or Debian people with out-of sync devel packages will get problems when tgoto expects terminfo-style caps
-         and their tgetstr returns termcap-style. This is a rather heavy-handed way of avoiding those problems: */    
-      DPRINTF0(DEBUG_TERMIO, "term_cursor_hpos appears to be termcap-style, not terminfo; NULLing it");
-      term_cursor_hpos = NULL;
-    }   
-    term_cursor_up      = my_tgetstr("up");
-    term_cursor_down    = my_tgetstr("do");
+    term_backspace      = tigetstr_or_else_tgetstr("cub1",   "le");
+    term_cr             = tigetstr_or_else_tgetstr("cr",     "cr");
+    term_clear_line     = tigetstr_or_else_tgetstr("el",     "ce");
+    term_clear_screen   = tigetstr_or_else_tgetstr("clear",  "cl");
+    term_cursor_hpos    = tigetstr_or_else_tgetstr("hpa",    "ch"); 
+    term_cursor_left    = tigetstr_or_else_tgetstr("cub1",   "le");
+    term_cursor_right   = tigetstr_or_else_tgetstr("cuf1",   "nd");
+    term_cursor_up      = tigetstr_or_else_tgetstr("cuu1",   "up");
+    term_cursor_down    = tigetstr_or_else_tgetstr("cud1",   "do");
+
+    
+    /* the following codes are never output by rlwrap, but used to filter out "garbage" that is coming from commands that use them */ 
+    term_rmcup          = tigetstr_or_else_tgetstr("rmcup",  "te"); 
+    term_rmkx           = tigetstr_or_else_tgetstr("rmkx",   "ke");
 
     /* 
-    setupterm(term_name, 1, (int *)0);
     term_colors         = tigetnum("colors");
     DPRINTF1(DEBUG_TERMIO, "terminal colors: %d", term_colors); 
     */
                                         
-  } 
+  }
+
+  if (!(term_cursor_hpos || term_cursor_right) || !term_cursor_up || !term_cursor_down)
+    myerror(WARNING|NOERRNO, "Your terminal '%s' is not fully functional, expect some problems.", term_name);
   term_eof = saved_terminal_settings.c_cc[VEOF];
   term_stop = saved_terminal_settings.c_cc[VSTOP];
 
@@ -163,7 +205,7 @@ set_echo(int yes)
  
   log_terminal_settings(pterm);
   if (tcsetattr(STDIN_FILENO, TCSANOW, pterm) < 0 && errno != ENOTTY)
-    ;   /* myerror(FATAL|USE_ERRNO, "cannot prepare terminal (tcsetattr error on stdin)"); */
+    { /* nothing */ }   /* myerror(FATAL|USE_ERRNO, "cannot prepare terminal (tcsetattr error on stdin)"); */
   free(pterm);
 }
 
@@ -172,11 +214,18 @@ int
 cursor_hpos(int col)
 {
   char *instantiated;
-  assert(term_cursor_hpos != NULL);     /* caller has to make sure */
-  instantiated = tgoto(term_cursor_hpos, 0, col); /* tgoto with a command that takes one parameter: parameter goes to 2nd arg ("vertical position"). */
-  assert(instantiated);
-  DPRINTF2(DEBUG_TERMIO, "tgoto(term_cursor_hpos, 0, %d) = %s", col, mangle_string_for_debug_log(instantiated, 20));
-  tputs(instantiated, 1, my_putchar);
+  if (term_cursor_hpos) {
+    instantiated = tgoto(term_cursor_hpos, 0, col); /* tgoto with a command that takes one parameter: parameter goes to 2nd arg ("vertical position"). */
+    assert(instantiated);
+    DPRINTF2(DEBUG_TERMIO, "tgoto(term_cursor_hpos, 0, %d) = %s", col, mangle_string_for_debug_log(instantiated, 20));
+    tputs(instantiated, 1, my_putchar);
+  } else {
+    int i;
+    assert(term_cursor_right != NULL);
+    cr();
+    for (i = 0; i < col; i++)
+      tputs(term_cursor_right, 1, my_putchar);
+  }
   return TRUE;
 }
 
@@ -250,15 +299,15 @@ curs_left()
 void
 curs_up()
 {
-  /* assert(term_cursor_up != NULL);    caller has to make sure */
-  tputs(term_cursor_up, 1, my_putchar);
+  if (term_cursor_up)
+    tputs(term_cursor_up, 1, my_putchar);
 }
 
 void
 curs_down()
 {
-  assert(term_cursor_down != NULL);     /* caller has to make sure */
-  tputs(term_cursor_down, 1, my_putchar);
+  if (term_cursor_down)
+    tputs(term_cursor_down, 1, my_putchar);
 }
 
 int my_putchar(TPUTS_PUTC_ARGTYPE c)
@@ -291,7 +340,7 @@ static void test_termfunc(char *control_string, char *control_string_name, char*
      mysavestring("NULL"));
   printf("\n%s = %s\n", control_string_name, mangled_control_string);
   if (!control_string)
-    return;
+    printf("trying without suitable control string, fasten seatbelts and brace for impact... \n");
   my_putstr(start);
   sleep(1);
   termfunc();
