@@ -13,7 +13,7 @@
     the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
     You may contact the author by:
-    e-mail:  hlub@knoware.nl
+    e-mail:  hanslub42@gmail.com
 */
 
 
@@ -23,7 +23,6 @@
 
 
 
-char slaves_working_directory[MAXPATHLEN+1]; 
 
 static FILE *log_fp;
 
@@ -123,7 +122,7 @@ my_pselect(int n, fd_set *readfds, fd_set *writefds, fd_set *exceptfds, const st
 }       
 
 struct termios *
-my_tcgetattr(int fd, char *which) /* the 'which' argument is only needed for error reporting */
+my_tcgetattr(int fd, char *UNUSED(which)) 
 {
   struct termios *pterm =
     (struct termios *)mymalloc(sizeof(struct termios));
@@ -131,10 +130,11 @@ my_tcgetattr(int fd, char *which) /* the 'which' argument is only needed for err
 }
 
 
+/* empty signal handler */
 void
-do_nothing(int unused)
+do_nothing(int UNUSED(signal))
 {
-  ; /* yawn.... */
+  /* yawn.... */
 }
 
 
@@ -305,10 +305,6 @@ void set_ulimit(int resource, long value) {
 
 
 
-
-
-
-
 int open_unique_tempfile(const char *suffix, char **tmpfile_name) {
   char **tmpdirs = list4(getenv("TMPDIR"), getenv("TMP"), getenv("TEMP"), "/tmp");
   char *tmpdir = first_of(tmpdirs);
@@ -329,7 +325,7 @@ int open_unique_tempfile(const char *suffix, char **tmpfile_name) {
   }
 #endif
   if (tmpfile_fd < 0)
-    myerror(FATAL|USE_ERRNO, "could not create temporary file %s", tmpfile_name);
+    myerror(FATAL|USE_ERRNO, "could not create readable/writable temporary file %s", tmpfile_name);
   free(tmpdirs);
   return tmpfile_fd;
 }  
@@ -402,6 +398,8 @@ myerror(int error_flags, const char *message_format, ...)
 }
 
 
+    
+
 void
 open_logfile(const char *filename)
 {
@@ -470,6 +468,28 @@ timestamp(char *buf, int size)
 }
 
 
+
+/* Dan Bernsteins djb2, hashing n strings in one go */
+unsigned long
+hash_multiple(int n, ...)
+{
+  unsigned long hash = 5381;
+  int i, c;
+  char *str;
+
+  va_list ap;
+  va_start(ap, n);
+  for(i = 0; i < n; i++) {
+    str = va_arg(ap, char *);
+    assert(str != NULL);
+    while ((c = (unsigned char) *str++))
+      hash = ((hash << 5) + hash) + c; /* hash * 33 + c */
+  }
+  va_end(ap);
+  return hash;
+}
+
+
 int killed_by(int status) {
 #ifdef WTERMSIG
   if (WIFSIGNALED(status))
@@ -480,36 +500,112 @@ int killed_by(int status) {
 
 /* change_working_directory() changes rlwrap's working directory to /proc/<command_pid>/cwd
    (on systems where this makes sense, like linux and Solaris) */
-   
+/* TODO: clean up this mess: 
+
+   * use only a global char *slaves_working_directory (not proc_pid_cwd)
+   * factor out a function int get_new_slave_cwd(char **cwd) that returns 1 (and updates slaves_working_directory) only 
+     if the working dir has changed, and use it like this:
+
+    void change_working_directory() { 
+         if(get_new_slave_cwd(&slaves_working_directory)) 
+           chdir(slaves_working_directory);
+    }  
+*/ 
+
+
+
+/* get_new_slave_cwd(&cwd)  finds the rlwrapped command's current working directory. If this differs from cwd, free(*cwd) and
+   set *cwd to (a copy of) the new working directory. Return value: 0, or 1 if rlwrap needs to do a chdir(cwd) to again have 
+   the same working dir as the rlwrapped command.
+*/
+static int
+get_new_slave_cwd(char **cwd)
+{
+  char *possibly_new_cwd = NULL;
+  int return_value = 0;
+
+  
+#if defined(HAVE_PROC_PID_CWD) /* Linux, Solaris, and FreeBSD with the proc filesystem */
+
+  static char *proc_pid_cwd = NULL;
+  char readlink_buffer[MAXPATHLEN+1];
+    
+  if (!proc_pid_cwd)
+    proc_pid_cwd = add3strings(PROC_MOUNTPOINT, "/", add2strings(as_string(command_pid), "/cwd"));
+
+# ifdef HAVE_READLINK
+  if  (readlink(proc_pid_cwd, readlink_buffer, MAXPATHLEN) > 0)
+    possibly_new_cwd = mysavestring(readlink_buffer);
+# else
+  /* readlink unavailable, use /proc/nnn/cwd ... */
+  possibly_new_cwd = mysavestring(proc_pid_cwd);
+  return_value = 1; /* always do a chdir(): the symlink /proc/nnn/cwd may point somewhere else now... */
+#  endif /* HAVE_READLINK */
+  
+#elif HAVE_DECL_PROC_PIDVNODEPATHINFO /* OS X */
+
+  struct proc_vnodepathinfo vpi;
+
+  if (proc_pidinfo(command_pid, PROC_PIDVNODEPATHINFO, 0, &vpi, sizeof(vpi)) <= 0)
+    DPRINTF2(DEBUG_COMPLETION, "proc_pidinfo(%d, .. failed: %s", command_pid, strerror(errno));
+  possibly_new_cwd = mysavestring(vpi.pvi_cdir.vip_path);
+
+#elif HAVE_FREEBSD_LIBPROCSTAT /* FreeBSD without the proc filesystem */
+
+  unsigned int count = 0;
+  struct procstat *procstat = procstat_open_sysctl();
+  struct kinfo_proc *kip = procstat_getprocs(procstat, KERN_PROC_PID, command_pid, &count);
+  struct filestat_list *head;
+  struct filestat *fst;
+  
+  if (count == 1) {
+    head = procstat_getfiles(procstat, kip, 0);
+  
+    STAILQ_FOREACH(fst, head, next) 
+      if (fst->fs_uflags & PS_FST_UFLAG_CDIR) {
+        possibly_new_cwd = mysavestring(fst->fs_path);
+        break;
+      }     
+    procstat_freefiles(procstat, head);
+    procstat_freeprocs(procstat, kip);
+    procstat_close(procstat);
+  }     
+
+#else
+  /* no HAVE_PROC_PID_CWD, HAVE_DECL_PROC_PIDVNODEPATHINFO or HAVE_FREEBSD_LIBPROCSTAT: do nothing */
+#endif
+  
+  if (possibly_new_cwd && strcmp(*cwd, possibly_new_cwd)) {
+    return_value = 1;
+    free(*cwd);
+    *cwd = mysavestring(possibly_new_cwd);
+  }
+  free(possibly_new_cwd);
+  return return_value;
+}     
+ 
+
+
+
 void
 change_working_directory()
 {
-#ifdef HAVE_PROC_PID_CWD
-  static char proc_pid_cwd[MAXPATHLEN+1];
-  static int initialized = FALSE;
-  int linklen = 0;
-
-  snprintf0(slaves_working_directory, MAXPATHLEN, "?");
-
-  if (!initialized && command_pid > 0) {        /* first time we're called after birth of child */
-    snprintf2(proc_pid_cwd, MAXPATHLEN , "%s/%d/cwd", PROC_MOUNTPOINT, command_pid);
-    initialized = TRUE;
-  }     
-  if (chdir(proc_pid_cwd) == 0) {
-#ifdef HAVE_READLINK
-    linklen = readlink(proc_pid_cwd, slaves_working_directory, MAXPATHLEN);
-    if (linklen > 0)
-      slaves_working_directory[linklen] = '\0';
-    else
-      snprintf1(slaves_working_directory, MAXPATHLEN, "? (%s)", strerror(errno));
-#endif /* HAVE_READLINK */
-  }                 
-#else /* HAVE_PROC_PID_CWD */
-  /* do nothing at all */
-#endif
-}
+  static char *slaves_working_directory = NULL;
+  if (!slaves_working_directory)
+    slaves_working_directory = mysavestring(".");
+  if(command_pid > 0 && get_new_slave_cwd(&slaves_working_directory)) {
+    if (chdir(slaves_working_directory)) {
+      DPRINTF2(DEBUG_COMPLETION, "chdir(%s) failed: %s", slaves_working_directory, strerror(errno));
+    } else {
+      DPRINTF1(DEBUG_COMPLETION, "chdir(%s): success", slaves_working_directory);
+    }   
+  }
+}       
 
 
+
+
+#undef isset
 #define isset(flag) ((flag) ? "set" : "unset")
 
 /* print info about terminal settings */
@@ -638,7 +734,7 @@ usage(int status)
   print_option('v', "version", NULL, FALSE, NULL);
   print_option('w', "wait-before-prompt", "N", FALSE, "(msec, <0  : patient mode)");
   print_option('W', "polling", NULL, FALSE, NULL);
-  print_option('z', "filter", "filter command", FALSE, "(-z listing lists installed filters)");  
+  print_option('z', "filter", "filter command", FALSE, "('rlwrap -z listing' writes a list of installed filters)");  
   
  
 #ifdef DEBUG
@@ -657,7 +753,7 @@ usage(int status)
   print_debug_flag (DEBUG_MEMORY_MANAGEMENT, "memory management,");
   print_debug_flag (DEBUG_FILTERING, "filtering,");
   print_debug_flag (DEBUG_COMPLETION, "and completion.");
-  /*  print_debug_flag (DEBUG_AD_HOC, "to see your own DEBUG_AD_HOC results"); */
+  print_debug_flag (DEBUG_AD_HOC, "to see your own DEBUG_AD_HOC results"); 
   print_debug_flag (DEBUG_WITH_TIMESTAMPS, "to add (relative) timestamps,");
   print_debug_flag (FORCE_HOMEGROWN_REDISPLAY, "to force the use of my_homegrown_redisplay(),");
   print_debug_flag (DEBUG_LONG_STRINGS, "to not limit the length of strings in debug log (sloooow!)");
@@ -671,7 +767,7 @@ usage(int status)
   fprintf(stderr,
           "\n"
           "bug reports, suggestions, updates:\n"
-          "http://utopia.knoware.nl/~hlub/uck/rlwrap/\n");
+          "https://github.com/hanslub42/rlwrap\n");
 
   exit(status);
 }           
@@ -695,12 +791,28 @@ mymalloc(size_t size)
   return ptr;
 }
 
-/* free of possibly NULL pointer (only needed on very olde systems) */
-void myfree(void *ptr) {
-  if (ptr)
-    free(ptr);
+
+/* free() with variable number of arguments. To show where the argumets end, the last argument should be special,
+   (and never a legitimate pointer) but we cannot use NULL (as the to-be-freed pointers may legitimately be NULL)
+   We now use FMEND (#defined as ((void *) -1) in rlwrap.h, but @@@ is this fool-proof?
+ */
+
+void
+free_multiple(void *ptr, ...)
+{
+  void *p;
+  va_list ap;
+  free(ptr);
+  va_start(ap, ptr);
+  while((p = va_arg(ap, void *)) != FMEND) {
+    free(p);
+  }
+  va_end(ap);
 }       
-  
+
+
+
+
 void mysetsid() {
 # ifdef HAVE_SETSID /* c'mon, this is POSIX! */
   pid_t ret = setsid();
@@ -727,10 +839,9 @@ static char *stored_cmdline_filename;
 
 
 
-void mirror_args_init(char**argv) {
-  int i;
+void mirror_args_init(char**argv) {  
 #ifdef ENABLE_MIRROR_ARGS
-
+  int i;
   rlwrap_command_argv = argv;
   stored_cmdline_filename = mymalloc(MAXPATHLEN);
   *stored_cmdline_filename = '\0';  
@@ -755,7 +866,7 @@ void mirror_args_init(char**argv) {
 
       
 
-void mirror_args(command_pid) {
+void mirror_args(pid_t command_pid) {
   int cmdline_fd;
   long cmdline_length;
   static int been_warned = 0;
